#include <stdio.h>
#include <wchar.h>
#include <windows.h>
#include "inc_xml/msxml2.h"
#include "inc/vss.h"
#include "inc/vswriter.h"
#include "inc/vsbackup.h"

static IVssBackupComponents* g_pVssBackup = 0;
static VSS_ID g_snapshotSetId;
static HINSTANCE g_handle = 0;
static bool g_comInitialized = false;

static const int MAX_VOLUME_LENGTH = 4;

extern "C"
{
#include "../rsync.h"

	int g_vssInitialized = 0;
	char g_vssVolume[MAX_PATH];
}

#define CHECKED_CALL(result) checked_call(result, __LINE__)

bool checked_call(int result, int line)
{
	if (FAILED(result))
	{
		rprintf(FERROR, "Error %d on line %d!\n", result, line);
		return false;
	}

	return true;
}

void format_volume(const char* path, wchar_t* volume)
{
	char tmp[MAX_VOLUME_LENGTH] = "";
	char tmp3[2] = "";
/*
	if (strlen (path) > )
*/
	/* /cygdrive/ == 10 */
	tmp3 [0] = path [10];
	/* atleast needs check on length */
	strcat(tmp3, ":\\");
	const char* tmp2 = tmp3;
	mbsrtowcs(volume, &tmp2, MAX_VOLUME_LENGTH, 0);
}

typedef long __stdcall CREATEVSSBACKUPCOMPONENTS(class IVssBackupComponents**);

extern "C" int initialize_vss(const char* path)
{
	if (g_vssInitialized)
		return false;

	g_handle = LoadLibrary("vssapi.dll");
	if (!g_handle)
	{
		printf("LoadLibrary failed on line %d!", __LINE__);
		return false;
	}
	
	CREATEVSSBACKUPCOMPONENTS* pCreateVssBackupComponents = reinterpret_cast<CREATEVSSBACKUPCOMPONENTS*>(
		GetProcAddress(g_handle, "?CreateVssBackupComponents@@YGJPAPAVIVssBackupComponents@@@Z"));
	if (!pCreateVssBackupComponents)
	{
		printf("GetProcAddress failed on line %d!", __LINE__);
		return false;
	}

	CHECKED_CALL(CoInitialize(0));
	g_comInitialized = true;

	CHECKED_CALL((*pCreateVssBackupComponents)(&g_pVssBackup));
	CHECKED_CALL(g_pVssBackup->InitializeForBackup());
	CHECKED_CALL(g_pVssBackup->StartSnapshotSet(&g_snapshotSetId));

	wchar_t volume[MAX_VOLUME_LENGTH];
	format_volume(path, volume);

	VSS_ID snapshotId;
	CHECKED_CALL(g_pVssBackup->AddToSnapshotSet(volume, GUID_NULL, &snapshotId));

	IVssAsync* pAsync = 0;
	CHECKED_CALL(g_pVssBackup->GatherWriterMetadata(&pAsync));
	CHECKED_CALL(pAsync->Wait());

	CHECKED_CALL(g_pVssBackup->SetBackupState(false, true, VSS_BT_FULL));
	CHECKED_CALL(g_pVssBackup->PrepareForBackup(&pAsync));
	CHECKED_CALL(pAsync->Wait());
	CHECKED_CALL(g_pVssBackup->DoSnapshotSet(&pAsync));
	CHECKED_CALL(pAsync->Wait());

	pAsync->Release();

	VSS_SNAPSHOT_PROP prop;
	CHECKED_CALL(g_pVssBackup->GetSnapshotProperties(snapshotId, &prop));
	sprintf(g_vssVolume, "%ls", prop.m_pwszSnapshotDeviceObject);

	rprintf (FINFO, "Using VSS volume: %s\n", g_vssVolume);

	g_vssInitialized = true;

	return true;
}

extern "C" void deinitialize_vss()
{
	if (g_pVssBackup)
	{
		long deleted = 0;
		VSS_ID nonDeletedId;
		CHECKED_CALL(g_pVssBackup->DeleteSnapshots(g_snapshotSetId, VSS_OBJECT_SNAPSHOT_SET, true, &deleted, &nonDeletedId));
		strcpy(g_vssVolume, "");
		
		g_pVssBackup->Release();
		g_pVssBackup = 0;
	}

	if (g_comInitialized)
	{
		CoUninitialize();
		g_comInitialized = false;
	}

	if (g_handle)
	{
		FreeLibrary(g_handle);
		g_handle = 0;
	}
}
